/* (C) 2022 YiRing, Inc. */
package com.yiring.app.service.user.impl;

import cn.hutool.core.collection.ListUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.github.liaochong.myexcel.core.DefaultStreamExcelBuilder;
import com.github.liaochong.myexcel.core.SaxExcelReader;
import com.yiring.app.domain.location.LocationTag;
import com.yiring.app.domain.location.LocationTagRepository;
import com.yiring.app.excel.user.UserExportExcel;
import com.yiring.app.excel.user.UserImportExcel;
import com.yiring.app.param.user.*;
import com.yiring.app.service.dept.DepartmentService;
import com.yiring.app.service.user.UserService;
import com.yiring.app.vo.user.UserVo;
import com.yiring.auth.domain.dept.Department;
import com.yiring.auth.domain.dept.DepartmentRepository;
import com.yiring.auth.domain.post.Post;
import com.yiring.auth.domain.post.PostRepository;
import com.yiring.auth.domain.user.User;
import com.yiring.auth.domain.user.UserRepository;
import com.yiring.common.core.Result;
import com.yiring.common.core.Status;
import com.yiring.common.domain.BasicEntity;
import com.yiring.common.param.IdParam;
import com.yiring.common.param.PageParam;
import com.yiring.common.vo.PageVo;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import javax.persistence.EntityManager;
import javax.persistence.criteria.*;
import javax.servlet.http.HttpServletResponse;
import lombok.extern.slf4j.Slf4j;
import org.apache.poi.ss.usermodel.Workbook;
import org.jetbrains.annotations.NotNull;
import org.springframework.data.domain.Example;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

/**
 * 用户
 * @author LJ-2204
 * @date 2022/4/20
 */

@Transactional(rollbackFor = RuntimeException.class)
@Service
@Slf4j
public class UserServiceImpl implements UserService {

    @Resource
    EntityManager em;

    @Resource
    UserRepository userRepository;

    @Resource
    LocationTagRepository locationTagRepository;

    @Resource
    DepartmentRepository departmentRepository;

    @Resource
    DepartmentService departmentService;

    @Resource
    PostRepository postRepository;

    @Override
    public Result<PageVo<UserVo>> findUserPage(UserFindParam param, PageParam pageParam) {
        CriteriaQuery<UserVo> cq = getUserVoCriteriaQuery(param);

        List<UserVo> userVos = em
            .createQuery(cq)
            .setFirstResult(pageParam.getPageNo())
            .setMaxResults(pageParam.getPageSize())
            .getResultList();
        PageVo<UserVo> page = PageVo.build(userVos, userVos.size());
        return Result.ok(page);
    }

    @Override
    public Result<UserVo> findUserById(IdParam idParam) {
        Optional<User> userOptional = userRepository.findOne(Example.of(User.builder().id(idParam.getId()).build()));
        if (userOptional.isPresent()) {
            UserVo userVo = UserVo.transformUserInfo(userOptional.get());
            return Result.ok(userVo);
        }
        return Result.no(Status.BAD_REQUEST);
    }

    @Override
    public Result<String> userBingTag(UserBingTagParam param) {
        Optional<LocationTag> locationTagOptional = locationTagRepository.findOne(
            Example.of(LocationTag.builder().code(param.getCode()).build())
        );
        if (locationTagOptional.isEmpty()) throw new RuntimeException("标签卡不存在");

        LocationTag locationTag = locationTagOptional.get();
        if (!ObjectUtil.isNotEmpty(locationTag.getUser())) {
            throw new RuntimeException("已绑定其他用户 : " + locationTag.getUser().getRealName());
        }

        locationTag.setUser(User.builder().id(param.getId()).build());
        locationTag.setUsed(true);
        locationTagRepository.save(locationTag);

        return Result.ok();
    }

    @Override
    public Result<String> userUnBingTag(IdParam idParam) {
        Optional<LocationTag> locationTagOptional = locationTagRepository.findOne(
            Example.of(LocationTag.builder().user(User.builder().id(idParam.getId()).build()).build())
        );
        if (locationTagOptional.isEmpty()) return Result.no(Status.BAD_REQUEST);

        LocationTag locationTag = locationTagOptional.get();
        locationTag.setUser(null);
        locationTag.setUsed(false);
        locationTagRepository.save(locationTag);

        return Result.ok();
    }

    @Override
    public Result<String> deleteUser(IdParam idParam) {
        LocationTag locationTagBuild = LocationTag.builder().user(User.builder().id(idParam.getId()).build()).build();
        long locationTagCount = locationTagRepository.count(Example.of(locationTagBuild));
        if (locationTagCount > 0) return Result.no(Status.BAD_REQUEST, "已发卡，不可删除");

        Department departmentBuild = Department.builder().leader(User.builder().id(idParam.getId()).build()).build();
        long departmentCount = departmentRepository.count(Example.of(departmentBuild));
        if (departmentCount > 0) return Result.no(Status.BAD_REQUEST, "部门负责人，请先解绑");

        userRepository.deleteById(idParam.getId());
        return Result.ok();
    }

    @Override
    public Result<PageVo<UserVo>> selectUser() {
        List<User> users = userRepository.findAll();
        List<UserVo> userVos = UserVo.transforms(users);
        PageVo<UserVo> pageVo = PageVo.build(userVos, userVos.size());
        return Result.ok(pageVo);
    }

    @Override
    public void exportUser(UserFindParam param, HttpServletResponse response) {
        CriteriaQuery<UserVo> cq = getUserVoCriteriaQuery(param);
        List<UserVo> userVos = em.createQuery(cq).getResultList();
        List<UserExportExcel> userExportExcels = userVos
            .stream()
            .map(UserExportExcel::transform)
            .collect(Collectors.toList());
        try (
            DefaultStreamExcelBuilder<UserExportExcel> streamExcelBuilder = DefaultStreamExcelBuilder
                .of(UserExportExcel.class)
                .threadPool(Executors.newFixedThreadPool(2))
                .rowHeight(14)
                .titleRowHeight(14)
                .style(
                    "cell->vertical-align:center;text-align:center",
                    "title->vertical-align:center;text-align:center;font-weight:bold;font-family:等线"
                )
                .start()
        ) {
            streamExcelBuilder.append(userExportExcels);

            String fileName = URLEncoder.encode("用户信息.xlsx", StandardCharsets.UTF_8);
            response.setContentType("application/octet-stream");
            response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
            response.setHeader("Content-Disposition", "attachment;filename=" + fileName);
            OutputStream out = response.getOutputStream();

            Workbook workbook = streamExcelBuilder.fixedTitles().build();
            workbook.write(out);
            workbook.close();
            out.flush();
            out.close();
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RuntimeException("用户信息导出失败: " + e.getMessage());
        }
    }

    @Override
    public Result<String> stateUser(IdParam idParam) {
        LocationTag locationTagBuild = LocationTag.builder().user(User.builder().id(idParam.getId()).build()).build();
        long locationTagCount = locationTagRepository.count(Example.of(locationTagBuild));
        if (locationTagCount > 0) return Result.no(Status.BAD_REQUEST, "已发卡，不可停用");

        Department departmentBuild = Department.builder().leader(User.builder().id(idParam.getId()).build()).build();
        long departmentCount = departmentRepository.count(Example.of(departmentBuild));
        if (departmentCount > 0) return Result.no(Status.BAD_REQUEST, "部门负责人，请先解绑");

        Optional<User> userOptional = userRepository.findOne(Example.of(User.builder().id(idParam.getId()).build()));
        if (userOptional.isEmpty()) return Result.no(Status.BAD_REQUEST);

        User user = userOptional.get();
        user.setEnabled(!user.getEnabled());

        userRepository.save(user);

        return Result.ok();
    }

    @Override
    public Result<String> modifyUser(UserModifyParam param) {
        Optional<User> userOptional = userRepository.findOne(Example.of(User.builder().id(param.getId()).build()));

        if (userOptional.isEmpty()) return Result.no(Status.BAD_REQUEST, "用户不存在");

        User user = userOptional.get();

        if (!StrUtil.equals(user.getUuid(), param.getUuid())) {
            long count = userRepository.count(Example.of(User.builder().uuid(param.getUuid()).build()));

            if (count > 0) return Result.no(Status.BAD_REQUEST, "工号相同");
        }

        User newUser = param.transform();

        userRepository.save(newUser);

        return Result.ok();
    }

    @Override
    public Result<String> addUser(UserAddParam param) {
        long count = userRepository.count(Example.of(User.builder().uuid(param.getUuid()).build()));

        if (count > 0) return Result.no(Status.BAD_REQUEST, "工号相同");

        User user = param.transform();
        userRepository.save(user);

        return Result.ok();
    }

    @Override
    public Result<String> importUser(MultipartFile file) {
        if (file.isEmpty()) return Result.no(Status.BAD_REQUEST, "请选择文件");
        List<UserImportExcel> userImportExcels;
        try (InputStream inputStream = file.getInputStream()) {
            userImportExcels =
                SaxExcelReader
                    .of(UserImportExcel.class)
                    .rowFilter(row -> row.getRowNum() > 0)
                    .ignoreBlankRow()
                    .read(inputStream);
        } catch (IOException e) {
            log.info(e.getMessage());
            throw new RuntimeException("文件导入异常");
        }
        // 部门信息
        HashMap<String, Department> hashMap = departmentService.getStringDepartmentHashMap();

        // 职位信息
        Map<String, List<Post>> postMap = postRepository
            .findAll()
            .stream()
            .collect(Collectors.groupingBy(Post::getName));

        userImportExcels =
            userImportExcels
                .stream()
                .filter(userImportExcel ->
                    userRepository.count(Example.of(User.builder().uuid(userImportExcel.getUuid()).build())) == 0
                )
                .filter(userImportExcel ->
                    locationTagRepository.count(
                        Example.of(LocationTag.builder().code(userImportExcel.getCode()).used(true).build())
                    ) ==
                    0
                )
                .collect(Collectors.toList());

        userImportExcels.forEach(userImportExcel -> {
            User user = UserImportExcel.transform(userImportExcel);
            Department department = hashMap.get(userImportExcel.getDeptName());
            Post post = postMap.get(userImportExcel.getPostName()).get(0);
            user.setDepartment(department);
            user.setPost(post);
            User saveUser = userRepository.save(user);
            if (StrUtil.isNotEmpty(userImportExcel.getCode())) {
                LocationTag locationTag = LocationTag.builder().code(userImportExcel.getCode()).user(saveUser).build();
                locationTagRepository.save(locationTag);
            }
        });

        return Result.ok();
    }

    @Override
    public Result<PageVo<UserVo>> indexUser(UserIndexParam param) {
        CriteriaBuilder cb = em.getCriteriaBuilder();
        CriteriaQuery<UserVo> cq = cb.createQuery(UserVo.class);
        Root<User> root = cq.from(User.class);
        Expression<String> realName = root.get(User.Fields.realName);
        Expression<String> uuid = root.get(User.Fields.uuid);
        Expression<User.Type> type = root.get(User.Fields.type);
        Expression<Boolean> gender = root.get(User.Fields.gender);
        Expression<Boolean> isSpecial = root.get(User.Fields.isSpecial);
        Expression<Department> department = root.get(User.Fields.department).get("name");
        Expression<String> mobile = root.get(User.Fields.mobile);
        Expression<Post> post = root.get(User.Fields.post).get("name");

        Subquery<String> query = cq.subquery(String.class);
        Root<LocationTag> tagRoot = query.from(LocationTag.class);
        query.select(tagRoot.get(LocationTag.Fields.code));
        query.where(cb.equal(tagRoot.get(LocationTag.Fields.user), root));

        cq.multiselect(realName, uuid, type, department, mobile, post, query, gender, isSpecial);
        List<Predicate> predicates = ListUtil.toList();

        if (StrUtil.isNotEmpty(param.getStr())) {
            if (StrUtil.isNumeric(param.getStr())) {
                predicates.add(cb.like(root.get(User.Fields.mobile), param.getStr() + "%"));
            } else {
                predicates.add(cb.like(root.get(User.Fields.realName), param.getStr() + "%"));
            }
        }

        Order order = cb.desc(root.get(BasicEntity.Fields.createTime));

        cq.orderBy(order).where(predicates.toArray(new Predicate[0])).getRestriction();

        List<UserVo> userVos = em.createQuery(cq).getResultList();
        PageVo<UserVo> page = PageVo.build(userVos, userVos.size());

        return Result.ok(page);
    }

    @NotNull
    private CriteriaQuery<UserVo> getUserVoCriteriaQuery(UserFindParam param) {
        CriteriaBuilder cb = em.getCriteriaBuilder();
        CriteriaQuery<UserVo> cq = cb.createQuery(UserVo.class);
        Root<User> root = cq.from(User.class);
        Expression<String> realName = root.get(User.Fields.realName);
        Expression<String> uuid = root.get(User.Fields.uuid);
        Expression<User.Type> type = root.get(User.Fields.type);
        Expression<Boolean> gender = root.get(User.Fields.gender);
        Expression<Boolean> isSpecial = root.get(User.Fields.isSpecial);
        Expression<Department> department = root.get(User.Fields.department).get("name");
        Expression<String> mobile = root.get(User.Fields.mobile);
        Expression<Post> post = root.get(User.Fields.post).get("name");

        Subquery<String> query = cq.subquery(String.class);
        Root<LocationTag> tagRoot = query.from(LocationTag.class);
        query.select(tagRoot.get(LocationTag.Fields.code));
        query.where(cb.equal(tagRoot.get(LocationTag.Fields.user), root));

        cq.multiselect(realName, uuid, type, department, mobile, post, query, gender, isSpecial);

        List<Predicate> predicates = ListUtil.toList();

        if (StrUtil.isNotEmpty(param.getRealName())) {
            predicates.add(cb.like(root.get(User.Fields.realName), "%" + param.getRealName() + "%"));
        }

        if (StrUtil.isNotEmpty(param.getCode())) {
            predicates.add(cb.like(root.get(LocationTag.Fields.code), "%" + param.getCode() + "%"));
        }

        if (StrUtil.isNotEmpty(param.getUuid())) {
            predicates.add(cb.like(root.get(User.Fields.uuid), "%" + param.getUuid() + "%"));
        }

        if (StrUtil.isNotEmpty(param.getMobile())) {
            predicates.add(cb.like(root.get(User.Fields.mobile), "%" + param.getMobile() + "%"));
        }

        if (ObjectUtil.isNotEmpty(param.getDeptId())) {
            predicates.add(cb.equal(root.get(User.Fields.department).get("id"), param.getDeptId()));
        }

        Order order = cb.desc(root.get(BasicEntity.Fields.createTime));

        cq.orderBy(order).where(predicates.toArray(new Predicate[0])).getRestriction();
        return cq;
    }
}
